
from idds.core.messages import retrieve_messages

"""
msgs = retrieve_messages(bulk=1000000)
to_check_msgs = []
for msg in msgs:
    # print(msg['msg_id'], msg)
    if msg['msg_content']['workload_id'] == 20525133:
        to_check_msgs.append(msg)
for to_check_msg in to_check_msgs:
    print(to_check_msg)
    contents = to_check_msg['msg_content']['files']
    print("number of files: %s" % len(contents))
    for content in contents:
        pass
        # print(content)
"""
msgs = retrieve_messages(bulk_size=1000000)
for msg in msgs:
    if msg['request_id'] == 152:
        print(msg)



#!/usr/bin/env python

from tabulate import tabulate

from idds.common.utils import json_dumps                 # noqa F401
from idds.common.constants import ContentStatus          # noqa F401
from idds.core.requests import get_requests              # noqa F401
from idds.core.messages import retrieve_messages         # noqa F401
from idds.core.transforms import get_transforms          # noqa F401
from idds.core.workprogress import get_workprogresses    # noqa F401
from idds.core.processings import get_processings        # noqa F401
from idds.core import transforms as core_transforms      # noqa F401


def show_req_transforms(request_id):
    reqs = get_requests(request_id=request_id, with_detail=False)
    trfs = get_transforms(request_id=request_id)
    prs = get_processings(request_id=request_id)
    reqs_times = {}
    trfs_times = {}
    prs_times = {}
    for req in reqs:
        reqs_times[req['request_id']] = {'created_at': req['created_at']}
    for trf in trfs:
        trfs_times[trf['transform_id']] = {'request_id': trf['request_id'], 'created_at': trf['created_at']}
    for pr in prs:
        prs_times[pr['processing_id']] = {'request_id': pr['request_id'], 'transform_id': pr['transform_id'], 'created_at': pr['created_at']}

    table = []
    title = ['request_id', 'req_created_at', 'transform_id', 'tf_created_at', 'processing_id', 'pr_created_at', 'request-processing']
    table.append(title)
    pr_ids = sorted(list(prs_times.keys()))
    for pr_id in pr_ids:
        req_id = prs_times[pr_id]['request_id']
        tf_id = prs_times[pr_id]['transform_id']
        row = [req_id, reqs_times[req_id]['created_at'], tf_id, trfs_times[tf_id]['created_at'], pr_id, prs_times[pr_id]['created_at'], (prs_times[pr_id]['created_at'] - reqs_times[req_id]['created_at']).seconds / 60]
        table.append(row)
    print(tabulate(table))


if __name__ == '__main__':
    req_ids = [i for i in range(100, 110)]
    # req_ids = [99]
    for req_id in req_ids:
        show_req_transforms(req_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


"""
Test hyper parameter optimization docker container in local env.
"""

import json
import os
import random
import re
import subprocess
import tempfile
# from uuid import uuid4 as uuid

# from idds.client.client import Client
# from idds.common.utils import replace_parameters_with_values, run_command
from idds.common.constants import RequestType, RequestStatus
# from idds.common.utils import get_rest_host


def get_test_codes():
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_codes = os.path.join(dir_name, 'activelearning_test_codes/activelearning_test_codes.tgz')
    return test_codes


def run_command(cmd):
    """
    Runs a command in an out-of-procees shell.
    """
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid)
    stdout, stderr = process.communicate()
    if stdout is not None and type(stdout) in [bytes]:
        stdout = stdout.decode()
    if stderr is not None and type(stderr) in [bytes]:
        stderr = stderr.decode()
    status = process.returncode
    return status, stdout, stderr


def replace_parameters_with_values(text, values):
    """
    Replace all strings starting with '%'. For example, for this string below, it should replace ['%NUM_POINTS', '%IN', '%OUT']
    'run --rm -it -v "$(pwd)":/payload gitlab-registry.cern.ch/zhangruihpc/endpointcontainer:latest /bin/bash -c "echo "--num_points %NUM_POINTS"; /bin/cat /payload/%IN>/payload/%OUT"'

    :param text
    :param values: parameter values, for example {'NUM_POINTS': 5, 'IN': 'input.json', 'OUT': 'output.json'}
    """
    for key in values:
        key1 = '%' + key
        text = re.sub(key1, str(values[key]), text)
    return text


def create_temp_dir():
    dirpath = tempfile.mkdtemp()
    return dirpath


def create_idds_input_json(job_dir, req_meta):
    input_json = 'idds_input_%s.json' % random.randint(1, 1000)
    if 'initial_points' in req_meta:
        points = req_meta['initial_points']
    else:
        points = []

    opt_space = None
    opt_points = {'points': points, 'opt_space': opt_space}
    if 'opt_space' in req_meta:
        opt_space = req_meta['opt_space']
        opt_points['opt_space'] = opt_space
    with open(os.path.join(job_dir, input_json), 'w') as f:
        json.dump(opt_points, f)
    return input_json


def get_output_json(req_meta):
    output_json = None
    if 'output_json' in req_meta:
        output_json = req_meta['output_json']
    return output_json


def get_exec(job_dir, req_meta):
    if 'max_points' in req_meta:
        max_points = req_meta['max_points']
    else:
        max_points = 20
    if 'num_points_per_generation' in req_meta:
        num_points = req_meta['num_points_per_generation']
    else:
        num_points = 10
    num_init_points = 0
    if 'initial_points' in req_meta:
        num_init_points = len(req_meta['initial_points'])
    num_points = min(max_points, num_points) - num_init_points

    idds_input_json = create_idds_input_json(job_dir, req_meta)
    output_json = get_output_json(req_meta)

    param_values = {'MAX_POINTS': max_points,
                    'NUM_POINTS': num_points,
                    'IN': idds_input_json,
                    'OUT': output_json}

    executable = req_meta['executable'].strip()
    arguments = req_meta['arguments']
    executable = replace_parameters_with_values(executable, param_values)
    arguments = replace_parameters_with_values(arguments, param_values)

    if executable == 'docker':
        executable = 'docker run --rm -v $(pwd):%s %s' % (req_meta['workdir'], req_meta['sandbox'])
    exect = executable + ' ' + arguments
    return exect


def get_req_properties():
    req_properties = {
        # 'scope': 'data15_13TeV',
        # 'name': 'data15_13TeV.00270949.pseudo.%s' % str(uuid()),
        'requester': 'panda',
        'request_type': RequestType.HyperParameterOpt,
        'transform_tag': 'prodsys2',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        # 'request_metadata': {'workload_id': '20525134', 'sandbox': None, 'executable': 'docker run --rm -it -v "$(pwd)":/payload gitlab-registry.cern.ch/zhangruihpc/endpointcontainer:latest /bin/bash -c "/bin/cat /payload/input_json.txt>/payload/output_json.txt"', 'arguments': '-s --input %IN', 'output_json': 'output.json'}  # noqa: E501
        # 'request_metadata': {'workload_id': '20525134', 'is_pseudo_input': True, 'sandbox': None, 'executable': 'docker', 'arguments': 'run --rm -it -v "$(pwd)":/payload gitlab-registry.cern.ch/zhangruihpc/endpointcontainer:latest /bin/cat /payload/%IN>%OUT', 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'output_json': 'output.json'}  # noqa: E501
        # 'request_metadata': {'workload_id': '20525134', 'is_pseudo_input': True, 'sandbox': None, 'method': 'nevergrad', 'opt_space': {'A': [1, 2, 3], 'B': (1, 10)}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 10}  # noqa: E501
        # 'request_metadata': {'workload_id': '20525134', 'sandbox': None, 'executable': 'docker', 'arguments': 'run -v $(pwd):/data wguanicedew/idds_hpo_nevergrad python /opt/hyperparameteropt_nevergrad.py --max_points=%MAX_POINTS --num_points=%NUM_POINTS --input=/data/%IN --output=/data/%OUT', 'output_json': 'output.json', 'opt_space': {"A": {"type": "Choice", "params": {"choices": [1, 4]}}, "B": {"type": "Scalar", "bounds": [0, 5]}}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 20, 'num_points_per_generation': 10}  # noqa: E501
        'request_metadata': {'workload_id': '20525134', 'sandbox': 'wguanicedew/idds_hpo_nevergrad', 'workdir': '/data', 'executable': 'docker', 'arguments': 'python /opt/hyperparameteropt_nevergrad.py --max_points=%MAX_POINTS --num_points=%NUM_POINTS --input=/data/%IN --output=/data/%OUT', 'output_json': 'output.json', 'opt_space': {"A": {"type": "Choice", "params": {"choices": [1, 4]}}, "B": {"type": "Scalar", "bounds": [0, 5]}}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 20, 'num_points_per_generation': 10}  # noqa: E501
    }
    return req_properties


def run_test():
    job_dir = create_temp_dir()
    print('job dir: %s' % job_dir)
    os.chdir(job_dir)
    print("$pwd: %s" % os.getcwd())
    req = get_req_properties()
    req_meta = req['request_metadata']
    exe = get_exec(job_dir, req_meta)
    print('executable: %s' % exe)
    status, output, error = run_command(exe)
    print('status: %s, output: %s, error: %s' % (status, output, error))
    output_json = get_output_json(req_meta)
    if not output_json:
        print('output_json is not defined')
    else:
        output_json = os.path.join(job_dir, output_json)
        if not os.path.exists(output_json):
            print("Failed: output_json is not created")
        else:
            try:
                with open(output_json, 'r') as f:
                    data = f.read()
                outputs = json.loads(data)
                print("outputs: %s" % str(outputs))
                if type(outputs) not in [list, tuple]:
                    print("Failed: outputs is not a list")
            except Exception as ex:
                print("Failed to parse output_json: %s" % ex)


if __name__ == '__main__':
    run_test()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023


"""
Utils to create the database or destroy the database
"""

import os
import traceback
from typing import Union

from alembic.config import Config
from alembic import command

import sqlalchemy
# from sqlalchemy.engine import reflection
from sqlalchemy.engine import Inspector
from sqlalchemy import inspect
from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.schema import CreateSchema, MetaData, Table, DropTable, ForeignKeyConstraint, DropConstraint
from sqlalchemy.sql.ddl import DropSchema

from idds.common.config import config_has_option, config_get
from idds.orm.base import session, models


def build_database(echo=True, tests=False):
    """Build the database. """
    engine = session.get_engine(echo=echo)

    if config_has_option('database', 'schema'):
        schema = config_get('database', 'schema')
        if schema and not engine.dialect.has_schema(engine, schema):
            print('Schema set in config, trying to create schema:', schema)
            try:
                engine.execute(CreateSchema(schema))
            except Exception as e:
                print('Cannot create schema, please validate manually if schema creation is needed, continuing:', e)
                print(traceback.format_exc())

    models.register_models(engine)

    # record the head version of alembic
    alembic_cfg_file = os.environ.get("ALEMBIC_CONFIG")
    alembic_cfg = Config(alembic_cfg_file)
    command.stamp(alembic_cfg, "head")


def destroy_database(echo=True):
    """Destroy the database"""
    engine = session.get_engine(echo=echo)

    try:
        models.unregister_models(engine)
    except Exception as e:
        print('Cannot destroy schema -- assuming already gone, continuing:', e)
        print(traceback.format_exc())


def destroy_everything(echo=True):
    """ Using metadata.reflect() to get all constraints and tables.
        metadata.drop_all() as it handles cyclical constraints between tables.
        Ref. http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DropEverything
    """
    engine = session.get_engine(echo=echo)

    try:
        # the transaction only applies if the DB supports
        # transactional DDL, i.e. Postgresql, MS SQL Server
        with engine.begin() as conn:

            inspector = inspect(conn)  # type: Union[Inspector, PGInspector]

            for tname, fkcs in reversed(
                    inspector.get_sorted_table_and_fkc_names(schema='*')):
                if tname:
                    drop_table_stmt = DropTable(Table(tname, MetaData(), schema='*'))
                    conn.execute(drop_table_stmt)
                elif fkcs:
                    if not engine.dialect.supports_alter:
                        continue
                    for tname, fkc in fkcs:
                        fk_constraint = ForeignKeyConstraint((), (), name=fkc)
                        Table(tname, MetaData(), fk_constraint)
                        drop_constraint_stmt = DropConstraint(fk_constraint)
                        conn.execute(drop_constraint_stmt)

            if config_has_option('database', 'schema'):
                schema = config_get('database', 'schema')
                if schema:
                    conn.execute(DropSchema(schema, cascade=True))

            if engine.dialect.name == 'postgresql':
                assert isinstance(inspector, PGInspector), 'expected a PGInspector'
                for enum in inspector.get_enums(schema='*'):
                    sqlalchemy.Enum(**enum).drop(bind=conn)

    except Exception as e:
        print('Cannot destroy db:', e)
        print(traceback.format_exc())


def dump_schema():
    """ Creates a schema dump to a specific database. """
    engine = session.get_dump_engine()
    models.register_models(engine)


def row2dict(row):
    """ Convert rows to dict. """
    row_dict = {}
    for col in row.keys():
        row_dict[str(col)] = row[col]
    return row_dict


def rows2dict(rows):
    """ Convert rows to dict. """
    results = []
    for row in rows:
        row_dict = {}
        for col in row.keys():
            row_dict[str(col)] = row[col]
        results.append(row_dict)
    return results



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


import heapq
import threading
import traceback
from concurrent import futures

from .timertask import TimerTask


class TimerScheduler(threading.Thread):
    """
    The base class to schedule Task which will be executed after some time
    """

    def __init__(self, num_threads, name=None, logger=None):
        super(TimerScheduler, self).__init__(name=name)
        self.num_threads = int(num_threads)
        if self.num_threads < 1:
            self.num_threads = 1
        self.graceful_stop = threading.Event()
        self.executors = futures.ThreadPoolExecutor(max_workers=self.num_threads,
                                                    thread_name_prefix=name)

        self._task_queue = []
        self._lock = threading.RLock()

        self.logger = logger

    def set_logger(self, logger):
        self.logger = logger

    def stop(self, signum=None, frame=None):
        self.graceful_stop.set()

    def create_task(self, task_func, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=10, priority=1):
        return TimerTask(task_func, task_output_queue, task_args, task_kwargs, delay_time, priority, self.logger)

    def add_task(self, task):
        with self._lock:
            heapq.heappush(self._task_queue, task)

    def remove_task(self, task):
        with self._lock:
            self._task_queue.remove(task)
            heapq.heapify(self._task_queue)

    def remove_all(self):
        with self._lock:
            self._task_queue = []

    def get_ready_task(self):
        with self._lock:
            if not self._task_queue:
                return None
            task = self._task_queue[0]
            if task.is_ready():
                heapq.heappop(self._task_queue)
                return task
        return None

    def execute_task(self, task):
        # self.logger.info('execute task: %s' % task)
        task.execute()
        self.add_task(task)

    def execute(self):
        while not self.graceful_stop.is_set():
            try:
                task = self.get_ready_task()
                if task:
                    self.executors.submit(self.execute_task, task)
                else:
                    self.graceful_stop.wait(1)
            except Exception as error:
                self.logger.critical("Caught an exception: %s\n%s" % (str(error), traceback.format_exc()))

    def execute_once(self):
        try:
            task = self.get_ready_task()
            if task:
                self.executors.submit(self.execute_task, task)
        except Exception as error:
            self.logger.critical("Caught an exception: %s\n%s" % (str(error), traceback.format_exc()))

    def execute_timer_schedule(self):
        try:
            task = self.get_ready_task()
            if task:
                self.executors.submit(self.execute_task, task)
        except Exception as error:
            self.logger.critical("Caught an exception: %s\n%s" % (str(error), traceback.format_exc()))



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

import copy

from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import is_new_version

from idds.workflow.work import Collection, Processing
from idds.workflow.workflow import Workflow

# from idds.atlas.workflow.atlasstageinwork import ATLASStageInWork
# from idds.atlas.workflow.atlashpowork import ATLASHPOWork


def convert_stagein_request_metadata_to_workflow(scope, name, workload_id, request_metadata):
    """
    Convert old format stagein request metadata of json to new format request metadata based on workflow.

    :param scope: The collection scope.
    :param name: The collection name.
    :param workload_id: The workload id.
    :param request_metadata: The request metadata.
    """
    # 'request_metadata': {'workload_id': '20776840', 'max_waiting_time': 3600, 'src_rse': 'NDGF-T1_DATATAPE', 'dest_rse': 'NDGF-T1_DATADISK', 'rule_id': '236e4bf87e11490291e3259b14724e30'}  # noqa: E501

    from idds.atlas.workflow.atlasstageinwork import ATLASStageinWork

    work = ATLASStageinWork(executable=None, arguments=None, parameters=None, setup=None,
                            exec_type='local', sandbox=None,
                            primary_input_collection={'scope': scope, 'name': name},
                            other_input_collections=None,
                            output_collections={'scope': scope, 'name': name},
                            log_collections=None,
                            logger=None,
                            max_waiting_time=request_metadata.get('max_waiting_time', 3600 * 7 * 24),
                            src_rse=request_metadata.get('src_rse', None),
                            dest_rse=request_metadata.get('dest_rse', None),
                            rule_id=request_metadata.get('rule_id', None))
    wf = Workflow()
    wf.set_workload_id(workload_id)
    wf.add_work(work)
    # work.set_workflow(wf)
    return wf


def convert_hpo_request_metadata_to_workflow(scope, name, workload_id, request_metadata):
    """
    Convert old format hpo request metadata of json to new format request metadata based on workflow.

    :param scope: The collection scope.
    :param name: The collection name.
    :param workload_id: The workload id.
    :param request_metadata: The request metadata.
    """
    # 'request_metadata': {'workload_id': '20525134', 'sandbox': None, 'method': 'bayesian', 'opt_space': {'A': (1, 4), 'B': (1, 10)}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 20, 'num_points_per_generation': 10}  # noqa: E501
    # 'request_metadata': {'workload_id': '20525135', 'sandbox': None, 'method': 'nevergrad', 'opt_space': {"A": {"type": "Choice", "params": {"choices": [1, 4]}}, "B": {"type": "Scalar", "bounds": [0, 5]}}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 20, 'num_points_per_generation': 10}  # noqa: E501
    # 'request_metadata': {'workload_id': '20525134', 'sandbox': 'wguanicedew/idds_hpo_nevergrad', 'workdir': '/data', 'executable': 'docker', 'arguments': 'python /opt/hyperparameteropt_nevergrad.py --max_points=%MAX_POINTS --num_points=%NUM_POINTS --input=/data/%IN --output=/data/%OUT', 'output_json': 'output.json', 'opt_space': {"A": {"type": "Choice", "params": {"choices": [1, 4]}}, "B": {"type": "Scalar", "bounds": [0, 5]}}, 'initial_points': [({'A': 1, 'B': 2}, 0.3), ({'A': 1, 'B': 3}, None)], 'max_points': 20, 'num_points_per_generation': 10}  # noqa: E501

    from idds.atlas.workflow.atlashpowork import ATLASHPOWork

    work = ATLASHPOWork(executable=request_metadata.get('executable', None),
                        arguments=request_metadata.get('arguments', None),
                        parameters=request_metadata.get('parameters', None),
                        setup=None, exec_type='local',
                        sandbox=request_metadata.get('sandbox', None),
                        method=request_metadata.get('method', None),
                        container_workdir=request_metadata.get('workdir', None),
                        output_json=request_metadata.get('output_json', None),
                        opt_space=request_metadata.get('opt_space', None),
                        workload_id=workload_id,
                        initial_points=request_metadata.get('initial_points', None),
                        max_points=request_metadata.get('max_points', None),
                        num_points_per_iteration=request_metadata.get('num_points_per_iteration', 10))
    wf = Workflow()
    wf.set_workload_id(workload_id)
    wf.add_work(work)
    return wf


def convert_old_workflow_to_new_workflow(data):
    if ('request_metadata' in data and data['request_metadata'] and 'workflow' in data['request_metadata']):
        workflow = data['request_metadata']['workflow']
        if workflow:
            for work_key in workflow.works_template:
                work = workflow.works_template[work_key]
                for coll_key in work.collections:
                    coll = work.collections[coll_key]
                    if type(coll) in [Collection]:
                        pass
                    else:
                        coll_metadata = copy.copy(coll)
                        del coll_metadata['scope']
                        del coll_metadata['name']
                        new_coll = Collection(scope=coll['scope'], name=coll['name'], coll_metadata=coll_metadata)
                        new_coll.internal_id = coll_key
                        work.collections[coll_key] = new_coll
                for proc_key in work.processings:
                    proc = work.processings[proc_key]
                    if type(proc) in [Processing]:
                        pass
                    else:
                        proc_metadata = proc['processing_metadata']
                        new_proc = Processing(processing_metadata=proc_metadata)
                        new_proc.internal_id = proc_key
                        if 'rule_id' in proc_metadata:
                            new_proc.external_id = proc_metadata['rule_id']
                        work.processings[proc_key] = new_proc
    return data


def convert_old_req_2_workflow_req(data):
    if not data:
        return data

    if data['request_type'] == RequestType.Workflow:
        if ('request_metadata' in data and data['request_metadata'] and 'version' in data['request_metadata']
            and data['request_metadata']['version'] and is_new_version(data['request_metadata']['version'], '0.2.9')):  # noqa W503
            return data
        else:
            data = convert_old_workflow_to_new_workflow(data)
            return data

    workload_id = None
    if 'workload_id' in data and data['workload_id']:
        workload_id = data['workload_id']
    elif 'workload_id' in data['request_metadata'] and data['request_metadata']['workload_id']:
        workload_id = data['request_metadata']['workload_id']

    if data['request_type'] in [RequestType.StageIn, RequestType.StageIn.value]:
        wf = convert_stagein_request_metadata_to_workflow(data['scope'], data['name'], workload_id,
                                                          data['request_metadata'])
        data['request_type'] = RequestType.Workflow
        data['transform_tag'] = 'workflow'
        data['status'] = RequestStatus.New
        data['workload_id'] = wf.get_workload_id()
        data['request_metadata'] = {'workload_id': wf.get_workload_id(),
                                    'workflow': wf}
        return data
    if data['request_type'] in [RequestType.HyperParameterOpt, RequestType.HyperParameterOpt.value]:
        wf = convert_hpo_request_metadata_to_workflow(data['scope'] if 'scope' in data else None,
                                                      data['name'] if 'name' in data else None,
                                                      workload_id,
                                                      data['request_metadata'])
        primary_init_work = wf.get_primary_initial_collection()
        if primary_init_work:
            if type(primary_init_work) in [dict]:
                data['scope'] = primary_init_work['scope']
                data['name'] = primary_init_work['name']
            elif type(primary_init_work) in [Collection]:
                data['scope'] = primary_init_work.scope
                data['name'] = primary_init_work.name

        data['request_type'] = RequestType.Workflow
        data['transform_tag'] = 'workflow'
        data['status'] = RequestStatus.New
        data['workload_id'] = wf.get_workload_id()
        data['request_metadata'] = {'workload_id': wf.get_workload_id(),
                                    'workflow': wf}
        return data
    return data


def convert_old_request_metadata(req):
    if 'request_metadata' in req and req['request_metadata']:
        wf = req['request_metadata']['workflow']
        req['request_metadata'] = wf
        return req
    return req



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
operations related to Messages.
"""

from idds.common.constants import MessageDestination, MessageType, MessageStatus
from idds.orm.base.session import read_session, transactional_session
from idds.orm import messages as orm_messages


@transactional_session
def add_message(msg_type, status, source, request_id, workload_id, transform_id,
                num_contents, msg_content, bulk_size=None, processing_id=0,
                destination=MessageDestination.Outside, session=None):
    """
    Add a message to be submitted asynchronously to a message broker.

    :param msg_type: The type of the msg as a number, e.g., finished_stagein.
    :param status: The status about the message
    :param source: The source where the message is from.
    :param msg_content: The message msg_content as JSON.
    :param session: The database session.
    """
    return orm_messages.add_message(msg_type=msg_type, status=status, source=source,
                                    request_id=request_id, workload_id=workload_id,
                                    transform_id=transform_id, num_contents=num_contents,
                                    destination=destination, processing_id=processing_id,
                                    bulk_size=bulk_size, msg_content=msg_content, session=session)


@transactional_session
def add_messages(messages, bulk_size=1000, session=None):
    return orm_messages.add_messages(messages, bulk_size=bulk_size, session=session)


@read_session
def retrieve_messages(bulk_size=None, msg_type=None, status=None, destination=None,
                      source=None, request_id=None, workload_id=None, transform_id=None,
                      processing_id=None, retries=None, delay=None, session=None):
    """
    Retrieve up to $bulk messages.

    :param bulk: Number of messages as an integer.
    :param msg_type: Return only specified msg_type.
    :param status: The status about the message
    :param source: The source where the message is from.
    :param session: The database session.

    :returns messages: List of dictionaries
    """
    return orm_messages.retrieve_messages(bulk_size=bulk_size, msg_type=msg_type,
                                          status=status, source=source, destination=destination,
                                          request_id=request_id, workload_id=workload_id,
                                          transform_id=transform_id, processing_id=processing_id,
                                          retries=retries, delay=delay,
                                          session=session)


@read_session
def retrieve_request_messages(request_id, bulk_size=1, session=None):
    return retrieve_messages(request_id=request_id,
                             msg_type=MessageType.IDDSCommunication,
                             status=MessageStatus.New,
                             bulk_size=bulk_size,
                             destination=MessageDestination.Clerk,
                             session=session)


@read_session
def retrieve_transform_messages(transform_id, bulk_size=1, session=None):
    return retrieve_messages(transform_id=transform_id,
                             msg_type=MessageType.IDDSCommunication,
                             status=MessageStatus.New,
                             bulk_size=bulk_size,
                             destination=MessageDestination.Transformer,
                             session=session)


@read_session
def retrieve_processing_messages(processing_id, bulk_size=1, session=None):
    return retrieve_messages(processing_id=processing_id,
                             msg_type=MessageType.IDDSCommunication,
                             status=MessageStatus.New,
                             bulk_size=bulk_size,
                             destination=MessageDestination.Carrier,
                             session=session)


@transactional_session
def delete_messages(messages, session=None):
    """
    Delete all messages with the given IDs.

    :param messages: The messages to delete as a list of dictionaries.
    """
    return orm_messages.delete_messages(messages=messages, session=session)


@transactional_session
def update_messages(messages, session=None):
    """
    Update all messages status with the given IDs.

    :param messages: The messages to be updated as a list of dictionaries.
    """
    return orm_messages.update_messages(messages=messages, session=session)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


from traceback import format_exc

from flask import Blueprint

from idds.common import exceptions
from idds.common.authentication import authenticate_is_super_user
from idds.common.constants import HTTP_STATUS_CODE
from idds.common.constants import RequestStatus
from idds.common.constants import (MessageType, MessageStatus,
                                   MessageSource, MessageDestination,
                                   CommandType)
from idds.common.utils import json_loads
from idds.core.requests import (add_request, get_requests,
                                get_request, update_request,
                                get_request_ids_by_name)
from idds.core.messages import add_message
from idds.core.commands import add_command
from idds.rest.v1.controller import IDDSController

from idds.rest.v1.utils import convert_old_req_2_workflow_req


class Requests(IDDSController):
    """ Get request """

    def get(self):
        """
        Get requests.

        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: A list containing requests.
        """

        try:
            request_id = self.get_request().args.get('request_id', None)
            workload_id = self.get_request().args.get('workload_id', None)
            if request_id is None and workload_id is None:
                self.generate_http_response(HTTP_STATUS_CODE.BadRequest,
                                            exc_cls=exceptions.BadRequest.__name__,
                                            exc_msg="request_id and workload_id are both None. One should not be None")
            # reqs = get_requests(request_id=request_id, workload_id=workload_id, to_json=True)
            reqs = get_requests(request_id=request_id, workload_id=workload_id)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=reqs)


class Request(IDDSController):
    """ Create, Update, get and delete Request. """

    def post(self):
        """ Create Request.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            500 Internal Error
        """
        try:
            parameters = self.get_request().data and json_loads(self.get_request().data)
            if 'status' not in parameters:
                parameters['status'] = RequestStatus.New
            if 'priority' not in parameters:
                parameters['priority'] = 0
            # if 'lifetime' not in parameters:
            #     parameters['lifetime'] = 30
            if 'username' not in parameters or not parameters['username']:
                if 'username' in self.get_request().environ and self.get_request().environ['username']:
                    parameters['username'] = self.get_request().environ['username']
        except ValueError:
            return self.generate_http_response(HTTP_STATUS_CODE.BadRequest, exc_cls=exceptions.BadRequest.__name__, exc_msg='Cannot decode json parameter dictionary')

        try:
            parameters = convert_old_req_2_workflow_req(parameters)
            request_id = add_request(**parameters)
        except exceptions.DuplicatedObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.Conflict, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data={'request_id': request_id})

    def put(self, request_id):
        """ Update Request properties with a given id.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            404 Not Found
            500 Internal Error
        """
        try:
            request = self.get_request()
            parameters = request.data and json_loads(request.data)
            # parameters['status'] = RequestStatus.Extend
        except ValueError:
            return self.generate_http_response(HTTP_STATUS_CODE.BadRequest, exc_cls=exceptions.BadRequest.__name__, exc_msg='Cannot decode json parameter dictionary')

        try:
            username = self.get_username()
            reqs = get_requests(request_id=request_id, with_request=True)
            for req in reqs:
                if req['username'] and req['username'] != username and not authenticate_is_super_user(username):
                    raise exceptions.AuthenticationNoPermission("User %s has no permission to update request %s" % (username, req['request_id']))
        except exceptions.AuthenticationNoPermission as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            # update_request(request_id, parameters)
            # msg = {'command': 'update_request', 'parameters': {'status': RequestStatus.ToSuspend}})
            msg = {'command': 'update_request', 'parameters': parameters}
            add_message(msg_type=MessageType.IDDSCommunication,
                        status=MessageStatus.New,
                        destination=MessageDestination.Clerk,
                        source=MessageSource.Rest,
                        request_id=request_id,
                        workload_id=None,
                        transform_id=None,
                        num_contents=1,
                        msg_content=msg)

        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data={'status': 0, 'message': 'update successfully'})

    def get(self, request_id, workload_id, with_detail, with_metadata=False, with_transform=False, with_processing=False):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None
            if with_detail and (type(with_detail) in [bool] or type(with_detail) in [str] and with_detail.lower() in ['true']):
                with_detail = True
            else:
                with_detail = False
            if with_metadata and (type(with_metadata) in [bool] or type(with_metadata) in [str] and with_metadata.lower() in ['true']):
                with_metadata = True
            else:
                with_metadata = False
            if with_transform and (type(with_transform) in [bool] or type(with_transform) in [str] and with_transform.lower() in ['true']):
                with_transform = True
            else:
                with_transform = False
            if with_processing and (type(with_processing) in [bool] or type(with_processing) in [str] and with_processing.lower() in ['true']):
                with_processing = True
            else:
                with_processing = False

            if with_detail or with_transform or with_processing:
                with_request = False
            else:
                with_request = True

            # reqs = get_requests(request_id=request_id, workload_id=workload_id, to_json=True)
            reqs = get_requests(request_id=request_id, workload_id=workload_id,
                                with_request=with_request, with_detail=with_detail,
                                with_metadata=with_metadata, with_transform=with_transform,
                                with_processing=with_processing)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=reqs)

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


class RequestName(IDDSController):
    """ Get id from name. """

    def get(self, name):
        """ Get id from name.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: {name:id} dict.
        """
        try:
            rets = get_request_ids_by_name(name)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)


class RequestBuild(IDDSController):
    """ Create, Update, get and delete Request. """

    def post(self, request_id):
        """ update build request result.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            500 Internal Error
        """
        try:
            parameters = self.get_request().data and json_loads(self.get_request().data)
            if 'signature' not in parameters or 'workflow' not in parameters:
                raise exceptions.IDDSException("signature and workflow are required")

            # request_id = parameters['request_id']
            signature = parameters['signature']
            workflow = parameters['workflow']

            req = get_request(request_id=request_id)
            if not req:
                raise exceptions.IDDSException("Request %s is not found" % request_id)
            if req['status'] not in [RequestStatus.Building]:
                raise exceptions.IDDSException("Request (request_id: %s, status: %s) is not in Building status" % (request_id, req['status']))

            build_workflow = req['request_metadata']['build_workflow']
            works = build_workflow.get_all_works()
            build_work = works[0]
            if build_work.get_signature() != signature:
                raise exceptions.IDDSException("Request (request_id: %s) has a different signature(%s != %s)" % (request_id,
                                                                                                                 signature,
                                                                                                                 build_work.get_signature()))
            req['request_metadata']['workflow'] = workflow

            parameters = {'status': RequestStatus.Built,
                          'request_metadata': req['request_metadata']}
            update_request(request_id=req['request_id'], parameters=parameters, update_request_metadata=True)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data={'request_id': request_id})


class RequestAbort(IDDSController):
    """ Abort Request. """

    def put(self, request_id, workload_id=None, task_id=None):
        """ Abort the request.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            404 Not Found
            500 Internal Error
        """
        if request_id == 'null':
            request_id = None
        if workload_id == 'null':
            workload_id = None
        if task_id == 'null':
            task_id = None

        try:
            username = self.get_username()
            if task_id:
                reqs = get_requests(request_id=request_id, workload_id=workload_id, with_processing=True)
            else:
                reqs = get_requests(request_id=request_id, workload_id=workload_id, with_request=True)

            if not reqs:
                return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(-1, {'status': -1, 'message': 'No match requests'})])
            matched_transform_id = None
            if task_id:
                for req in reqs:
                    if str(req['processing_workload_id']) == str(task_id):
                        matched_transform_id = req['transform_id']
                if matched_transform_id:
                    return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(-1, {'status': -1, 'message': 'No match tasks'})])

            for req in reqs:
                if req['username'] and req['username'] != username and not authenticate_is_super_user(username):
                    msg = "User %s has no permission to update request %s" % (username, req['request_id'])
                    # raise exceptions.AuthenticationNoPermission(msg)
                    return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(-1, {'status': -1, 'message': msg})])
        except exceptions.AuthenticationNoPermission as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            cmd_content = None
            if task_id and matched_transform_id:
                cmd_content = {'task_id': task_id,
                               'transform_id': matched_transform_id}

            add_command(request_id=request_id, cmd_type=CommandType.AbortRequest,
                        workload_id=workload_id, cmd_content=cmd_content,
                        username=username)

        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(0, {'status': 0, 'message': 'Command registered successfully'})])


class RequestRetry(IDDSController):
    """ Retry Request. """

    def put(self, request_id, workload_id=None):
        """ Retry the request.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            404 Not Found
            500 Internal Error
        """

        if request_id == 'null':
            request_id = None
        if workload_id == 'null':
            workload_id = None

        try:
            username = self.get_username()
            reqs = get_requests(request_id=request_id, workload_id=workload_id, with_request=True)
            if not reqs:
                return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(-1, {'status': -1, 'message': 'No match requests'})])

            for req in reqs:
                if req['username'] and req['username'] != username and not authenticate_is_super_user(username):
                    msg = "User %s has no permission to update request %s" % (username, req['request_id'])
                    # raise exceptions.AuthenticationNoPermission(msg)
                    return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(-1, {'status': -1, 'message': msg})])
        except exceptions.AuthenticationNoPermission as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            add_command(request_id=request_id, cmd_type=CommandType.ResumeRequest,
                        workload_id=workload_id, cmd_content=None,
                        username=username)

        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=[(0, {'status': 0, 'message': 'Command registered successfully'})])


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('request', __name__)

    request_view = Request.as_view('request')
    bp.add_url_rule('/request', view_func=request_view, methods=['post', ])
    bp.add_url_rule('/request/<request_id>', view_func=request_view, methods=['put', ])
    bp.add_url_rule('/request/<request_id>/<workload_id>/<with_detail>', view_func=request_view, methods=['get', ])
    bp.add_url_rule('/request/<request_id>/<workload_id>/<with_detail>/<with_metadata>', view_func=request_view, methods=['get', ])
    bp.add_url_rule('/request/<request_id>/<workload_id>/<with_detail>/<with_metadata>/<with_transform>', view_func=request_view, methods=['get', ])
    bp.add_url_rule('/request/<request_id>/<workload_id>/<with_detail>/<with_metadata>/<with_transform>/<with_processing>', view_func=request_view, methods=['get', ])

    request_name2id = RequestName.as_view('request_name')
    bp.add_url_rule('/request/name/<name>', view_func=request_name2id, methods=['get', ])

    request_build = RequestBuild.as_view('request_build')
    bp.add_url_rule('/request/build/<request_id>', view_func=request_build, methods=['post', ])

    request_abort = RequestAbort.as_view('request_abort')
    bp.add_url_rule('/request/abort/<request_id>/<workload_id>', view_func=request_abort, methods=['put', ])
    bp.add_url_rule('/request/abort/<request_id>/<workload_id>/task_id', view_func=request_abort, methods=['put', ])

    request_retry = RequestRetry.as_view('request_retry')
    bp.add_url_rule('/request/retry/<request_id>/<workload_id>', view_func=request_retry, methods=['put', ])

    return bp



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2021 - 2022


from flask import Blueprint

# from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE
from idds.rest.v1.controller import IDDSController


class Ping(IDDSController):
    """ Ping the rest service """

    def get(self):
        """ Ping the rest service.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        rets = {'Status': 'OK'}
        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('ping', __name__)

    view = Ping.as_view('ping')
    bp.add_url_rule('/ping', view_func=view, methods=['get'])
    return bp



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023

import traceback

from idds.common import exceptions
from idds.common.constants import ProcessingStatus, ProcessingLocking
from idds.common.utils import setup_logging, truncate_string
from idds.core import catalog as core_catalog
from idds.core import processings as core_processings
from idds.agents.common.eventbus.event import (EventType,
                                               UpdateTransformEvent,
                                               TriggerProcessingEvent,
                                               TerminatedProcessingEvent,
                                               SyncProcessingEvent)

from .utils import handle_trigger_processing, is_process_terminated
from .poller import Poller

setup_logging(__name__)


class Trigger(Poller):
    """
    Trigger works to trigger to release jobs
    """

    def __init__(self, num_threads=1, poll_period=10, retries=3, retrieve_bulk_size=2,
                 name='Trigger', message_bulk_size=1000, **kwargs):
        self.set_max_workers()
        if hasattr(self, 'trigger_max_number_workers'):
            self.max_number_workers = int(self.trigger_max_number_workers)

        num_threads = self.max_number_workers * 2
        super(Trigger, self).__init__(num_threads=num_threads, name=name, **kwargs)

        if hasattr(self, 'trigger_max_number_workers'):
            self.max_number_workers = int(self.trigger_max_number_workers)
        self.number_msg_workers = 0

    def is_ok_to_run_more_msg_processings(self):
        if self.number_msg_workers >= self.max_number_workers:
            return False
        return True

    def get_trigger_processings(self):
        """
        Get trigger processing
        """
        try:
            self.show_queue_size()
            if not self.is_ok_to_run_more_processings():
                return []
            # self.show_queue_size()

            processing_status = [ProcessingStatus.ToTrigger, ProcessingStatus.Triggering]
            processings = core_processings.get_processings_by_status(status=processing_status,
                                                                     locking=True, update_poll=True,
                                                                     not_lock=True,
                                                                     only_return_id=True,
                                                                     bulk_size=self.retrieve_bulk_size)
            if processings:
                self.logger.info("Main thread get [ToTrigger, Triggering] processings to process: %s" % (str(processings)))

            for pr_id in processings:
                self.logger.info("UpdateProcessingEvent(processing_id: %s)" % pr_id)
                event = TriggerProcessingEvent(publisher_id=self.id, processing_id=pr_id)
                self.event_bus.send(event)

            return processings
        except exceptions.DatabaseException as ex:
            if 'ORA-00060' in str(ex):
                self.logger.warn("(cx_Oracle.DatabaseError) ORA-00060: deadlock detected while waiting for resource")
            else:
                # raise ex
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return []

    def handle_trigger_processing(self, processing, trigger_new_updates=False):
        try:
            log_prefix = self.get_log_prefix(processing)
            ret_trigger_processing = handle_trigger_processing(processing,
                                                               self.agent_attributes,
                                                               trigger_new_updates=trigger_new_updates,
                                                               logger=self.logger,
                                                               log_prefix=log_prefix)
            process_status, update_contents, ret_msgs, parameters, update_dep_contents_status_name, update_dep_contents_status, new_update_contents, ret_update_transforms = ret_trigger_processing

            self.logger.debug(log_prefix + "handle_trigger_processing: ret_update_transforms: %s" % str(ret_update_transforms))

            new_process_status = process_status
            if is_process_terminated(process_status):
                new_process_status = ProcessingStatus.Terminating

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': new_process_status,
                                                'substatus': process_status,
                                                'locking': ProcessingLocking.Idle}}

            if parameters:
                # special parameters such as 'output_metadata'
                for p in parameters:
                    update_processing['parameters'][p] = parameters[p]

            ret = {'update_processing': update_processing,
                   'update_contents': update_contents,
                   'messages': ret_msgs,
                   'new_update_contents': new_update_contents,
                   'update_transforms': ret_update_transforms,
                   'update_dep_contents': (processing['request_id'], update_dep_contents_status_name, update_dep_contents_status),
                   'processing_status': new_process_status}
        except exceptions.ProcessFormatNotSupported as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())

            retries = processing['update_retries'] + 1
            if not processing['max_update_retries'] or retries < processing['max_update_retries']:
                proc_status = ProcessingStatus.Running
            else:
                proc_status = ProcessingStatus.Failed
            error = {'update_err': {'msg': truncate_string('%s' % (ex), length=200)}}

            # increase poll period
            update_poll_period = int(processing['update_poll_period'].total_seconds() * self.poll_period_increase_rate)
            if update_poll_period > self.max_update_poll_period:
                update_poll_period = self.max_update_poll_period

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': proc_status,
                                                'locking': ProcessingLocking.Idle,
                                                'update_retries': retries,
                                                'update_poll_period': update_poll_period,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)

            ret = {'update_processing': update_processing,
                   'update_contents': []}
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())

            retries = processing['update_retries'] + 1
            if not processing['max_update_retries'] or retries < processing['max_update_retries']:
                proc_status = ProcessingStatus.Running
            else:
                proc_status = ProcessingStatus.Failed
            error = {'update_err': {'msg': truncate_string('%s' % (ex), length=200)}}
            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': proc_status,
                                                'locking': ProcessingLocking.Idle,
                                                'update_retries': retries,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)
            update_processing['parameters'] = self.load_poll_period(processing, update_processing['parameters'])

            ret = {'update_processing': update_processing,
                   'update_contents': []}
        return ret

    def process_trigger_processing_real(self, event):
        try:
            if event:
                original_event = event
                # pr_status = [ProcessingStatus.New]
                self.logger.info("process_trigger_processing, event: %s" % str(event))
                pr = self.get_processing(processing_id=event._processing_id, status=None, locking=True)
                if not pr:
                    self.logger.error("Cannot find processing for event: %s" % str(event))
                else:
                    log_pre = self.get_log_prefix(pr)
                    self.logger.info(log_pre + "process_trigger_processing")
                    ret = self.handle_trigger_processing(pr)
                    # self.logger.info(log_pre + "process_trigger_processing result: %s" % str(ret))

                    new_update_contents = ret.get('new_update_contents', None)
                    ret['new_update_contents'] = None
                    ret_update_contents = ret.get('update_contents', None)
                    self.update_processing(ret, pr)

                    if new_update_contents or ret_update_contents:
                        # self.logger.info(log_pre + "update_contents_to_others_by_dep_id")
                        # core_catalog.update_contents_to_others_by_dep_id(request_id=pr['request_id'], transform_id=pr['transform_id'])
                        # self.logger.info(log_pre + "update_contents_to_others_by_dep_id done")

                        # core_catalog.delete_contents_update(request_id=pr['request_id'], transform_id=pr['transform_id'])
                        update_transforms = core_catalog.get_updated_transforms_by_content_status(request_id=pr['request_id'],
                                                                                                  transform_id=pr['transform_id'])
                        self.logger.info(log_pre + "update_transforms: %s" % str(update_transforms))
                        for update_transform in update_transforms:
                            if 'transform_id' in update_transform:
                                update_transform_id = update_transform['transform_id']
                                if update_transform_id != pr['transform_id']:
                                    event = UpdateTransformEvent(publisher_id=self.id,
                                                                 transform_id=update_transform_id,
                                                                 content={'event': 'Trigger'})
                                    self.logger.info(log_pre + "Trigger UpdateTransformEvent(transform_id: %s)" % update_transform_id)
                                    self.event_bus.send(event)

                    if (('processing_status' in ret and ret['processing_status'] == ProcessingStatus.Terminating)
                        or (event._content and 'Terminated' in event._content and event._content['Terminated'])):   # noqa W503
                        self.logger.info(log_pre + "TerminatedProcessingEvent(processing_id: %s)" % pr['processing_id'])
                        event = TerminatedProcessingEvent(publisher_id=self.id,
                                                          processing_id=pr['processing_id'],
                                                          content=event._content,
                                                          counter=original_event._counter)
                        self.event_bus.send(event)
                    else:
                        if ((event._content and 'has_updates' in event._content and event._content['has_updates'])
                            or ('update_contents' in ret and ret['update_contents'])    # noqa W503
                            or ('new_contents' in ret and ret['new_contents'])          # noqa W503
                            or ('messages' in ret and ret['messages'])):                # noqa E129
                            self.logger.info(log_pre + "SyncProcessingEvent(processing_id: %s)" % pr['processing_id'])
                            event = SyncProcessingEvent(publisher_id=self.id, processing_id=pr['processing_id'],
                                                        counter=original_event._counter)
                            self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())

    def process_trigger_processing(self, event):
        self.number_workers += 1
        self.process_trigger_processing_real(event)
        self.number_workers -= 1

    def process_msg_trigger_processing(self, event):
        self.number_msg_workers += 1
        self.process_trigger_processing_real(event)
        self.number_msg_workers -= 1

    def init_event_function_map(self):
        self.event_func_map = {
            EventType.TriggerProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_trigger_processing
            },
            EventType.MsgTriggerProcessing: {
                'pre_check': self.is_ok_to_run_more_msg_processings,
                'exec_func': self.process_msg_trigger_processing
            }
        }

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")

            self.load_plugins()
            self.init()

            self.add_default_tasks()

            self.init_event_function_map()

            task = self.create_task(task_func=self.get_trigger_processings, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=60, priority=1)
            self.add_task(task)

            self.execute()
        except KeyboardInterrupt:
            self.stop()


if __name__ == '__main__':
    agent = Trigger()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


"""
core operations related to workflow model.
"""

# from idds.common import exceptions
from idds.common.constants import WorkprogressStatus, WorkprogressLocking
from idds.orm.base.session import read_session, transactional_session
from idds.orm import workprogress as orm_workprogress, transforms as orm_transforms
from idds.workflowv2.work import WorkStatus


def create_workprogress(request_id, workload_id, scope, name, priority=0, status=WorkprogressStatus.New,
                        locking=WorkprogressLocking.Idle,
                        expired_at=None, errors=None, workprogress_metadata=None, processing_metadata=None):
    """
    Create a workprogress.

    :param request_id: The request id.
    :param workload_id: The workload id.
    :param scope: The scope.
    :param name: The name.
    :param status: The status as integer.
    :param locking: The locking as integer.
    :param priority: The priority as integer.
    :param expired_at: The datetime when the workprogress will be expired at.
    :param errors: The errors as a json.
    :param workprogress_metadata: The metadata as json.
    :param processing_metadata: The metadata as json.

    :returns: workprogress.
    """
    return orm_workprogress.create_workprogress(request_id=request_id, workload_id=workload_id, scope=scope, name=name,
                                                priority=priority, status=status,
                                                locking=locking, expired_at=expired_at,
                                                workprogress_metadata=workprogress_metadata,
                                                processing_metadata=processing_metadata)


@transactional_session
def add_workprogress(request_id, workload_id, scope, name, priority=0, status=WorkprogressStatus.New,
                     locking=WorkprogressLocking.Idle,
                     expired_at=None, errors=None, workprogress_metadata=None, processing_metadata=None,
                     session=None):
    """
    Add a workprogress.

    :param request_id: The request id.
    :param workload_id: The workload id.
    :param scope: The scope.
    :param name: The name.
    :param status: The status as integer.
    :param locking: The locking as integer.
    :param priority: The priority as integer.
    :param expired_at: The datetime when the workprogress will be expired at.
    :param errors: The errors as a json.
    :param workprogress_metadata: The metadata as json.
    :param processing_metadata: The metadata as json.

    :raises DuplicatedObject: If a workprogress with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: workprogress id.
    """

    return orm_workprogress.add_workprogress(request_id=request_id, workload_id=workload_id,
                                             scope=scope, name=name, priority=priority, status=status,
                                             locking=locking, expired_at=expired_at,
                                             workprogress_metadata=workprogress_metadata,
                                             processing_metadata=processing_metadata,
                                             session=session)


@transactional_session
def add_workprogresses(workprogresses, bulk_size=1000, session=None):
    """
    Add workprogresses.

    :param workprogresses: dict of workprogress.
    :param session: session.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: workprogress ids.
    """
    return orm_workprogress.add_workprogresses(workprogresses, bulk_size=bulk_size, session=session)


@read_session
def get_workprogresses(request_id=None, to_json=False, session=None):
    """
    Get workprogresses with request_id.

    :param request_id: The request_id of the request.
    :param to_json: Whether to return json format.
    :param session: The database session in use.

    :raises NoObject: If no workprogress is founded.

    :returns: list of workprogresses.
    """

    return orm_workprogress.get_workprogresses(request_id=request_id, to_json=to_json, session=session)


@read_session
def get_workprogress(workprogress_id, to_json=False, session=None):
    """
    Get a workprogress or raise a NoObject exception.

    :param workprogress_id: The id of the workprogress.
    :param to_json: whether to return json format.

    :param session: The database session in use.

    :raises NoObject: If no workprogress is founded.

    :returns: Workprogress.
    """
    return orm_workprogress.get_workprogress(workprogress_id=workprogress_id, to_json=to_json, session=session)


@transactional_session
def get_workprogresses_by_status(status, period=None, locking=False, bulk_size=None, to_json=False, session=None):
    """
    Get workprogresses.

    :param status: list of status of the workprogress data.
    :param locking: Wheter to lock workprogresses to avoid others get the same workprogress.
    :param bulk_size: Size limitation per retrieve.
    :param to_json: whether to return json format.

    :raises NoObject: If no workprogresses are founded.

    :returns: list of Workprogress.
    """

    return orm_workprogress.get_workprogresses_by_status(status=status, period=period, locking=locking,
                                                         bulk_size=bulk_size, to_json=to_json, session=session)


@transactional_session
def update_workprogress(workprogress_id, parameters, new_transforms=None, update_transforms=None, session=None):
    """
    update a workprogress.

    :param workprogress_id: the workprogress id.
    :param parameters: A dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no workprogress is founded.
    :raises DatabaseException: If there is a database error.

    """

    if new_transforms:
        for tf in new_transforms:
            orginal_work = tf['transform_metadata']['orginal_work']
            del tf['transform_metadata']['orginal_work']
            tf_id = orm_transforms.add_transform(**tf, session=session)
            # work = tf['transform_metadata']['work']
            orginal_work.set_work_id(tf_id, transforming=True)
            orginal_work.set_status(WorkStatus.New)
    if update_transforms:
        for tr_id in update_transforms:
            orm_transforms.update_transform(transform_id=tr_id, parameters=update_transforms[tr_id], session=session)
    return orm_workprogress.update_workprogress(workprogress_id=workprogress_id, parameters=parameters, session=session)


@transactional_session
def delete_workprogress(workprogress_id=None, session=None):
    """
    delete a workprogress.

    :param workprogress_id: The id of the workprogress.
    :param session: The database session in use.

    :raises NoObject: If no workprogress is founded.
    :raises DatabaseException: If there is a database error.
    """
    return orm_workprogress.update_workprogress(workprogress_id=workprogress_id, session=session)


@transactional_session
def clean_locking(time_period=3600, session=None):
    """
    Clean locking which is older than time period.

    :param time_period in seconds
    """
    return orm_workprogress.clean_locking(time_period=time_period, session=session)


@transactional_session
def clean_next_poll_at(status, session=None):
    """
    Clean next_poll_at.

    :param status: status of the workprogress
    """
    return orm_workprogress.clean_next_poll_at(status=status, session=session)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022


import traceback

from flask import Blueprint

from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE, CollectionRelationType
from idds.core.catalog import get_collections, get_contents, get_contents_ext, combine_contents_ext
from idds.rest.v1.controller import IDDSController


class Collections(IDDSController):
    """ Catalog """

    def get(self, scope, name, request_id, workload_id, relation_type):
        """ Get collections by scope, name, request_id and workload_id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if scope in ['null', 'None']:
                scope = None
            if name in ['null', 'None']:
                name = None
            if request_id in ['null', 'None']:
                request_id = None
            else:
                request_id = int(request_id)
            if workload_id in ['null', 'None']:
                workload_id = None
            else:
                workload_id = int(workload_id)
            if relation_type in ['null', 'None']:
                relation_type = None
            else:
                relation_type = int(relation_type)

            rets = get_collections(scope=scope, name=name, request_id=request_id, workload_id=workload_id,
                                   relation_type=relation_type, to_json=False)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(traceback.format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)


class Contents(IDDSController):
    """ Catalog """

    def get(self, coll_scope, coll_name, request_id, workload_id, relation_type, status):
        """ Get contents by coll_scope, coll_name, request_id, workload_id and relation_type.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: contents.
        """

        try:
            if coll_scope in ['null', 'None']:
                coll_scope = None
            if coll_name in ['null', 'None']:
                coll_name = None
            if request_id in ['null', 'None']:
                request_id = None
            else:
                request_id = int(request_id)
            if workload_id in ['null', 'None']:
                workload_id = None
            else:
                workload_id = int(workload_id)
            if relation_type in ['null', 'None']:
                relation_type = None
            else:
                relation_type = int(relation_type)
            if status in ['null', 'None']:
                status = None
            else:
                status = int(status)

            rets = get_contents(coll_scope=coll_scope, coll_name=coll_name, request_id=request_id,
                                workload_id=workload_id, relation_type=relation_type, status=status, to_json=False)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(traceback.format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)


class ContentsOutputExt(IDDSController):
    """ Catalog """

    def get(self, request_id, workload_id, transform_id, group_by_jedi_task_id=False):
        """ Get contents by request_id, workload_id and transform_id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: contents.
        """

        try:
            if request_id in ['null', 'None']:
                request_id = None
            else:
                request_id = int(request_id)
            if workload_id in ['null', 'None']:
                workload_id = None
            else:
                workload_id = int(workload_id)
            if transform_id in ['null', 'None']:
                transform_id = None
            else:
                transform_id = int(transform_id)
            if group_by_jedi_task_id:
                if type(group_by_jedi_task_id) in [bool]:
                    pass
                else:
                    if type(group_by_jedi_task_id) in [str] and group_by_jedi_task_id.lower() in ['true']:
                        group_by_jedi_task_id = True
                    else:
                        group_by_jedi_task_id = False
            else:
                group_by_jedi_task_id = False

            if request_id is None:
                return self.generate_http_response(HTTP_STATUS_CODE.BadRequest,
                                                   exc_cls=exceptions.BadRequest.__name__,
                                                   exc_msg="request_id must not be None")

            else:
                contents = get_contents(request_id=request_id, workload_id=workload_id, transform_id=transform_id,
                                        relation_type=CollectionRelationType.Output)
                contents_ext = get_contents_ext(request_id=request_id, workload_id=workload_id, transform_id=transform_id)

                ret_contents = combine_contents_ext(contents, contents_ext, with_status_name=True)
                rets = {}
                for content in ret_contents:
                    if group_by_jedi_task_id:
                        jedi_task_id = content.get('jedi_task_id', 'None')
                        if jedi_task_id not in rets:
                            rets[jedi_task_id] = []
                        rets[jedi_task_id].append(content)
                    else:
                        transform_id = content.get('transform_id')
                        if transform_id not in rets:
                            rets[transform_id] = []
                        rets[transform_id].append(content)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(traceback.format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('catalog', __name__)

    # catalog_view = Catalog.as_view('catalog')

    collections_view = Collections.as_view('collections')
    bp.add_url_rule('/catalog/collections/<scope>/<name>/<request_id>/<workload_id>/<relation_type>',
                    view_func=collections_view, methods=['get', ])  # get collections

    contents_view = Contents.as_view('contents')
    bp.add_url_rule('/catalog/contents/<coll_scope>/<coll_name>/<request_id>/<workload_id>/<relation_type>/<status>',
                    view_func=contents_view, methods=['get', ])  # get contents

    contents_ext_view = ContentsOutputExt.as_view('contents_output_ext')
    bp.add_url_rule('/catalog/contents_output_ext/<request_id>/<workload_id>/<transform_id>/<group_by_jedi_task_id>',
                    view_func=contents_ext_view, methods=['get', ])

    return bp





class TestProperty:

    def __init__(self):
        # @self.IDDSProperty
        pass

    def IDDSProperty(attribute):
        def _get(self, attribute):
            self.get_metadata_item(attribute, None)

        def _set(self, attribute, value):
            self.add_metadata_item(attribute, value)

        attribute = property(_get, _set)
        return attribute



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022

import json
from traceback import format_exc

from flask import Blueprint

from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE
from idds.common.constants import CollectionRelationType, ContentStatus
from idds.core import catalog
from idds.rest.v1.controller import IDDSController


class HyperParameterOpt(IDDSController):
    """  get and update hyper parameters. """

    def put(self, workload_id, request_id, id, loss):
        """ Update the loss for the hyper parameter.

        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            404 Not Found
            500 Internal Error
        """
        try:
            if workload_id == 'null':
                workload_id = None
            if request_id == 'null':
                request_id = None

            if workload_id is None and request_id is None:
                error = "One of workload_id and request_id should not be None or empty"
                return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            contents = catalog.get_contents(request_id=request_id, workload_id=workload_id,
                                            relation_type=CollectionRelationType.Output)

            if id:
                new_contents = []
                for content in contents:
                    if str(content['name']) == str(id):
                        new_contents.append(content)
                contents = new_contents
            content = contents[0]

            loss = float(loss)
            content_id = content['content_id']
            point = content['path']
            param, origin_loss = json.loads(point)
            params = {'path': json.dumps((param, loss)),
                      'status': ContentStatus.Available,
                      'substatus': ContentStatus.Available}
            catalog.update_content(content_id, params)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data={'status': 0, 'message': 'update successfully'})

    def get(self, workload_id, request_id, id=None, status=None, limit=None):
        """ Get hyper parameters.
        :param request_id: The id of the request.
        :param status: status of the hyper parameters. None for all statuses.
        :param limit: Limit number of hyperparameters.

        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: list of hyper parameters.
        """

        try:
            if workload_id == 'null':
                workload_id = None
            if request_id == 'null':
                request_id = None

            if status == 'null':
                status = None
            if limit == 'null':
                limit = None
            if id == 'null':
                id = None

            contents = catalog.get_contents(request_id=request_id, workload_id=workload_id,
                                            status=status, relation_type=CollectionRelationType.Output)

            if id:
                new_contents = []
                for content in contents:
                    if str(content['name']) == str(id):
                        new_contents.append(content)
                contents = new_contents

            if contents and limit and len(contents) > limit:
                contents = contents[:limit]

            hyperparameters = []
            for content in contents:
                point = content['path']
                parameter, loss = json.loads(point)
                param = {'id': content['name'],
                         'parameters': parameter,
                         'loss': loss}
                hyperparameters.append(param)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=hyperparameters)

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('hpo', __name__)

    hpo_view = HyperParameterOpt.as_view('hpo')
    bp.add_url_rule('/hpo/<workload_id>/<request_id>/<id>/<loss>', view_func=hpo_view, methods=['put', ])
    bp.add_url_rule('/hpo/<workload_id>/<request_id>/<id>/<status>/<limit>', view_func=hpo_view, methods=['get', ])
    return bp
